/*
 *  <Short Description>
 *  Copyright (C) 2023  Brett Terpstra
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <blt/gfx/model.h>
#include <blt/std/memory_util.h>

namespace blt::gfx
{
    void static_dynamic_array::swap()
    {
        size_t next_size = blt::mem::next_byte_allocation(size_);
        
        auto* new_data = new vbo_t[next_size];
        
        if (std::holds_alternative<vbo_t*>(data_))
        {
            auto* ptr = std::get<vbo_t*>(data_);
            std::memcpy(new_data, ptr, size_);
            delete[] ptr;
        } else {
            auto data = std::get<array_t>(data_);
            std::memcpy(new_data, data.data(), DATA_SIZE);
        }
        data_ = new_data;
        size_ = next_size;
    }
    
    static_dynamic_array::static_dynamic_array()
    {
        auto ptr = std::get<array_t>(data_);
        std::memset(ptr.data(), 0, DATA_SIZE);
    }
    
    /*
     * -----------------------------------
     *              vbo_t
     * -----------------------------------
     */
    void vbo_t::create()
    {
        glGenBuffers(1, &bufferID_);
    }
    
    void vbo_t::destroy()
    {
        glDeleteBuffers(1, &bufferID_);
    }
    
    void vbo_t::allocate(GLsizeiptr size, GLint buffer_type, GLint memory_type, void* data)
    {
        size_ = size;
        glBufferData(buffer_type, size, data, memory_type);
    }
    
    void vbo_t::update(GLsizeiptr size, void* data, GLint buffer_type, GLint memory_type)
    {
        if (size <= size_)
            sub_update(0, size, data, buffer_type);
        else
            allocate(size, buffer_type, memory_type, data);
        size_ = size;
    }
    
    void vbo_t::sub_update(GLsizeiptr offset, GLsizeiptr size, void* data, GLint buffer_type)
    {
        glBufferSubData(buffer_type, offset, size, data);
    }
    
    void vbo_t::bind(GLint buffer_type) const
    {
        glBindBuffer(buffer_type, bufferID_);
    }
    
    /*
     * ----------------------------
     *      basic_vertex_array
     * ----------------------------
     */
    
    basic_vertex_array::basic_vertex_array(): vaoID(0)
    {
        glGenVertexArrays(1, &vaoID);
    }
}