/*
 *  <Short Description>
 *  Copyright (C) 2023  Brett Terpstra
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#define STB_IMAGE_IMPLEMENTATION
#define STB_IMAGE_RESIZE_IMPLEMENTATION
#define STB_IMAGE_WRITE_IMPLEMENTATION
#define STB_PERLIN_IMPLEMENTATION

#include <blt/gfx/stb/stb_image_write.h>
#if defined(__GNUC__) || defined(__GNUG__)
    #pragma GCC diagnostic push
    #pragma GCC diagnostic ignored "-Warray-bounds"
    #include <blt/gfx/stb/stb_image_resize2.h>
    #pragma GCC diagnostic pop
#else
    #include <blt/gfx/stb/stb_image_resize2.h>
#endif
#include <blt/gfx/stb/stb_image.h>
#include <blt/gfx/stb/stb_perlin.h>
#include <blt/gfx/texture.h>

blt::gfx::texture_file::texture_file(const std::string& path, const std::string& name): m_name(name.empty() ? path : name), m_path(path)
{
    m_texture.m_data = stbi_load(
            m_path.c_str(),
            reinterpret_cast<int*>(&m_texture.m_width), reinterpret_cast<int*>(&m_texture.m_height),
            reinterpret_cast<int*>(&m_texture.m_channels),
            0
    );
}

blt::gfx::texture_file& blt::gfx::texture_file::resize(int target_width, int target_height)
{
    if (target_width == m_texture.width() && target_height == m_texture.height())
        return *this;
    // since we will be replacing the loaded data pointer, is it wise to use the allocator
    // that matches with what stb image uses, which is malloc, since we unload with stbi_image_free -> (free)
    auto* output_Data = (unsigned char*) malloc(
            target_width * target_height * m_texture.channels()
    );
    
    // resize the texture
    if (stbir_resize_uint8_linear(
            // input
            m_texture.m_data, m_texture.width(), m_texture.height(), 0,
            // output
            output_Data, target_width, target_height, 0,
            // channels
            static_cast<stbir_pixel_layout>(m_texture.channels())
    ))
    {
        BLT_WARN("Error resizing block texture image!");
    }
    
    // free up the old data
    stbi_image_free(m_texture.m_data);
    m_texture.m_data = output_Data;
    m_texture.m_width = target_width;
    m_texture.m_height = target_height;
    
    return *this;
}

void blt::gfx::texture_gl::setDefaults() const
{
    glTexParameteri(textureBindType, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(textureBindType, GL_TEXTURE_WRAP_T, GL_REPEAT);
    // nearest preserves the pixely look
    glTexParameteri(textureBindType, GL_TEXTURE_MIN_FILTER, GL_NEAREST_MIPMAP_LINEAR);
    glTexParameteri(textureBindType, GL_TEXTURE_MAG_FILTER, GL_NEAREST_MIPMAP_LINEAR);
#ifdef GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT
    // Anisotropy helps preserve textures at oblique angles
                float a = 0;
                glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &a);
                glTexParameterf(textureBindType, GL_TEXTURE_MAX_ANISOTROPY_EXT, a);
#endif
}
